/* 
 * EDOMI beta (Version 1.56) - Intelligente Steuerung und Visualisierung KNX-basierter Elektro-Installationen.
 * @link http://www.edomi.de 
 * @author Dr. Christian Gärtner <dr.christian.gaertner@googlemail.com> 
 * @copyright 2015,2016,2017 Dr. Christian Gärtner 
 * @license Freeware 
 *
 *     This file and all other files are part of EDOMI.
 *
 *     EDOMI is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 *     Diese und alle anderen Dateien sind Teil von EDOMI.
 *
 *     EDOMI wird in der Hoffnung, dass es nützlich sein wird, aber
 *     OHNE JEDE GEWÄHRLEISTUNG, bereitgestellt; sogar ohne die implizite
 *     Gewährleistung der MARKTFÄHIGKEIT oder EIGNUNG FÜR EINEN BESTIMMTEN ZWECK.
*/ 
function app2_init(winId) {

	//--- ggf. anpassen ---
	app2_zoomFaktor=1;									//Default-Zoomfaktor (beliebig definierbar)
	app2_el_shadowSelected="0px 0px 3px 5px "+apps_colorSelected;	//Schatten eines Elements, wenn markiert
	app2_previewValue='';								//Default: KO-Wert für Vorschau
	app2_elementOutlineColor="#ffff00";							//Rahmenfarbe für Elemente
	app2_elementOutlineColorDisabled="rgba(255,255,0,0.25)";	//Rahmenfarbe für geschützte Elemente (bzw. Inklude-Elemente)
	//---------------------

	app2_winId=winId;
	app2_pageId=0;
	app2_bufferSelected=null;			//Zwischenspeicher für Selektion
	app2_bufferGroups=null;				//Zwischenspeicher für Gruppen-Zustand
	app2_elementSelectRectData={mode:0,x1:0,y1:0,x2:0,y2:0};	//Auswahlrahmen

	//Verschieben eines Elements
	app2_dragElements=null;				//Drag-Elemente: Array[0..oo] = object{obj,objId,left,top,width,height,offsetX,offsetY}
	app2_dragElement=null;				//das "angefasste" Element beim Draggen
	app2_dragMoved=false;				//true=beim Draggen wurden Elemente bewegt, false=es wurde nur geklickt
	app2_dragKeyboardX=0;				//Position (Mausersatz)
	app2_dragKeyboardY=0;				//Position (Mausersatz)

	//WinId beim Verschieben per Tastatur
	app2_elementsKeyboardMoveWinId='';

	//Raster
	app2_gridXmode=true;
	app2_gridYmode=true;

	//Default-Zoom einstellen
	app2_zoom(app2_zoomFaktor);

	//Vorschau
	app2_previewMode=false;		
	
	app2_layerMode=0;						//aktueller Layer-Modus
}

function app2_zoom(n) {
	app2_zoomFaktor=n;
	var d=document.getElementById(app2_winId+"-page");
	d.style.webkitTransform="scale("+app2_zoomFaktor+")";
	d.style.width=((1/app2_zoomFaktor)*100)+"%";
	d.style.height=((1/app2_zoomFaktor)*100)+"%";
	document.getElementById(app2_winId+"-zoom").innerHTML=parseInt(n*100)+"%";
}

function app2_quit(winId) {
	app2_winId=null;
	closeWindow(winId);
}

function app2_refreshAll(pageId,list) {
	//Alle Elemente in "page" komplett löschen (wird von app2.php aufgerufen)
	//wird bei JEDEM Neuaufbau der Elemente aufgerufen (z.B. nach dem Löschen/Erstellen/Verbinden/...)
	//pageId: 0=aktuelle Seite refreshen, 1..x=diese Seite aufrufen
	//list: (optinal) String mit den IDs der Elemente, die beim Aufbau markiert sein sollen (für Duplizierung): z.B. "1;2;3;4;5;"
	if (pageId==0) {
		pageId=app2_pageId;
		app2_saveState();
	} else {
		app2_bufferSelected=new Array();
		app2_bufferGroups=new Array();
	}

	if (!list) {list="";}

	app2_pageId=pageId;
	ajax("start",2,app2_winId,pageId+AJAX_SEPARATOR1+((app2_previewMode)?1:0)+AJAX_SEPARATOR1+app2_previewValue,list);
}

function app2_saveState() {
	app2_bufferSelected=new Array();
	app2_bufferGroups=new Array();
	var visupage=document.getElementById(app2_winId+"-visupage");
	if (visupage && app2_pageId>0) {
		var element=visupage.querySelectorAll("[data-elementselected='1']");
		for (var t=0; t<element.length; t++) {app2_bufferSelected.push(element[t].id);}
		var element=document.getElementById(app2_winId+"-menu").querySelectorAll("[data-groupexpanded='1']");
		for (var t=0; t<element.length; t++) {app2_bufferGroups.push(element[t].id);}
	}
}

function app2_restoreState(list) {
	app2_elementSelectNone();
	if (list) {
		var element=list.split(";");
		for (var t=0; t<element.length; t++) {app2_elementSelect(app2_winId+"-element-"+element[t],1);}
	} else {
		for (var t=0; t<app2_bufferSelected.length; t++) {app2_elementSelect(app2_bufferSelected[t],1);}
	}
	for (var t=0; t<app2_bufferGroups.length; t++) {app2_groupCollapse(app2_bufferGroups[t],1);}
	app2_bufferSelected=new Array();
	app2_bufferGroups=new Array();
}



/*
============================================================================
GUI-Events
============================================================================
*/

//---------------------------------------------
//Element verschieben
//---------------------------------------------

function app2_elementDragStart(objId,keyMode) {
	if (!event) var event=window.event;

	app2_dragElements=new Array();
	app2_dragMoved=false;
	var visuPage=document.getElementById(app2_winId+"-visupage");

	if (!keyMode) {
		var mouseX=event.pageX*(1/app2_zoomFaktor);
		var mouseY=event.pageY*(1/app2_zoomFaktor);
	} else {
		var mouseX=0;
		var mouseY=0;	
		app2_dragKeyboardX=0;
		app2_dragKeyboardY=0;
	}

	//Anfasser-Element merken
	app2_dragElement=document.getElementById(objId);

	if (!keyMode) {
		//Anfasser temporär selektieren
		app2_saveState();
		app2_elementSelect(objId,1);	//Anfasser-Element markieren (ggf. auch das Gruppenelement selbst)
		if (app2_dragElement.dataset.controltyp==255) {
			app2_groupSelect(objId,1); 	//Anfasser-Gruppe komplett markieren
		}
	}

	//Boundingbox der selektierten Elemente ermitteln und merken
	var bBox=app2_selectionBoundingBox();
	app2_dragElements.push({
		obj:null,
		left:bBox.x,
		top:bBox.y,
		width:bBox.width,
		height:bBox.height,
		offsetX:parseInt(mouseX-bBox.x),
		offsetY:parseInt(mouseY-bBox.y)
	});

	//Selektierte Elemente merken (einschl. des Anfassers): offsetX/Y=Position relativ zur Boundingbox(!)
	var elements=visuPage.querySelectorAll("[data-elementselected='1']");
	for (var t=0; t<elements.length; t++) {
		var obj=document.getElementById(elements[t].id);
		app2_dragElements.push({
			obj:obj,
			left:obj.style.left.replace("px",""),
			top:obj.style.top.replace("px",""),
			offsetX:parseInt(obj.style.left.replace("px","")-bBox.x),
			offsetY:parseInt(obj.style.top.replace("px","")-bBox.y)
		}); 
	}

	//Koordinaten anzeigen (Boundingbox)
	document.getElementById(app2_winId+"-position").innerHTML=parseInt(app2_dragElements[0].left)+" / "+parseInt(app2_dragElements[0].top);

	//Fadenkreuz (Boundingbox)
	app2_CrosshairX=app2_newDiv(visuPage,app2_winId+"-crosshairX");
	app2_CrosshairX.className="app2_CrosshairX";
	app2_CrosshairX.style.top=app2_dragElements[0].top+"px";
	app2_CrosshairX.style.height=app2_dragElements[0].height+"px";
	app2_CrosshairX.style.width=visuPage.scrollWidth+"px";
	app2_CrosshairY=app2_newDiv(visuPage,app2_winId+"-crosshairY");
	app2_CrosshairY.className="app2_CrosshairY";
	app2_CrosshairY.style.left=app2_dragElements[0].left+"px";
	app2_CrosshairY.style.width=app2_dragElements[0].width+"px";
	app2_CrosshairY.style.height=visuPage.scrollHeight+"px";

	if (!keyMode) {
		window.addEventListener("mousemove",app2_elementDragMove,false);
		window.addEventListener("mouseup",app2_elementDragEnd,false);
	}
}

function app2_elementDragMove(event,keyMode,dX,dY) {

	if (!event) var event=window.event;

	app2_dragMoved=true;

	var visuPage=document.getElementById(app2_winId+"-visupage");
	var visupageWidth=visuPage.style.width.replace("px","");
	var visupageHeight=visuPage.style.height.replace("px","");
	
	//Raster einstellen
	var rasterX=1;
	var rasterY=1;
	if (app2_gridXmode) {rasterX=parseInt(document.getElementById(app2_winId+"-rasterx").value);}
	if (app2_gridYmode) {rasterY=parseInt(document.getElementById(app2_winId+"-rastery").value);}

	if (!keyMode) {
		var mouseX=event.pageX*(1/app2_zoomFaktor);
		var mouseY=event.pageY*(1/app2_zoomFaktor);
	} else {
		app2_dragKeyboardX+=dX*rasterX;
		app2_dragKeyboardY+=dY*rasterY;
		var mouseX=app2_dragKeyboardX;
		var mouseY=app2_dragKeyboardY;	
	}

	//Boundingbox positionieren (mit Raster und Grenzen)
	//neue Koordinaten
	var newX=parseInt(mouseX-app2_dragElements[0].offsetX);
	var newY=parseInt(mouseY-app2_dragElements[0].offsetY);
	//Raster anwenden
	newX=Math.floor(newX/rasterX)*rasterX;
	newY=Math.floor(newY/rasterY)*rasterY;
	//Grenzen prüfen
	if (newX<0) {newX=0;}
	if (newY<0) {newY=0;}
	if (newX>(visupageWidth-app2_dragElements[0].width)) {newX=visupageWidth-app2_dragElements[0].width;}
	if (newY>(visupageHeight-app2_dragElements[0].height)) {newY=visupageHeight-app2_dragElements[0].height;}
	app2_dragElements[0].left=newX;
	app2_dragElements[0].top=newY;

	//Elemente neu positionieren (relativ zur Boundingbox)
	for (var t=1; t<app2_dragElements.length; t++) {	//t=1 (ohne Boundingbox!)
		app2_dragElements[t].left=parseInt(newX+app2_dragElements[t].offsetX);
		app2_dragElements[t].top=parseInt(newY+app2_dragElements[t].offsetY);
		app2_dragElements[t].obj.style.left=parseInt(newX+app2_dragElements[t].offsetX)+"px";
		app2_dragElements[t].obj.style.top=parseInt(newY+app2_dragElements[t].offsetY)+"px";

	}

	//Koordinaten anzeigen (Boundingbox)
	document.getElementById(app2_winId+"-position").innerHTML=parseInt(app2_dragElements[0].left)+" / "+parseInt(app2_dragElements[0].top);

	//Fadenkreuz (Boundingbox)
	app2_CrosshairY.style.left=app2_dragElements[0].left+"px";
	app2_CrosshairX.style.top=app2_dragElements[0].top+"px";
	app2_CrosshairX.style.width=visuPage.scrollWidth+"px";
	app2_CrosshairY.style.height=visuPage.scrollHeight+"px";

	app2_groupBoundingBoxes();
}

function app2_elementDragEnd(event,keyMode,keySave) {
	if (!event) var event=window.event;
	if (app2_dragMoved && (!keyMode || keySave)) {
		//Positionen speichern
		var data="";
		for (var t=1; t<app2_dragElements.length; t++) {	//t=1 (ohne Boundingbox!)
			data+=app2_dragElements[t].obj.dataset.elementid+",";
			data+=parseInt(app2_dragElements[t].left-app2_dragElements[t].obj.dataset.translatex)+",";
			data+=parseInt(app2_dragElements[t].top-app2_dragElements[t].obj.dataset.translatey)+";";
		}
		ajax("saveElementsPosition",2,app2_winId,app2_pageId,data);
		app2_groupBoundingBoxes();
	}

	if (!keyMode) {
		app2_restoreState();
	}

	//Aufräumen
	app2_dragElements.length=0;

	if (!keyMode) {
		window.removeEventListener("mousemove",app2_elementDragMove,false);
		window.removeEventListener("mouseup",app2_elementDragEnd,false);
	}
	
	app2_CrosshairX.parentNode.removeChild(app2_CrosshairX);
	app2_CrosshairY.parentNode.removeChild(app2_CrosshairY);
}



//---------------------------------------------
//Element/Gruppe angeklickt
//---------------------------------------------
function app2_elementClick(objId) {
	if (!event) var event=window.event;
	if (event.button==0) {app2_elementClickLeft(objId);}
	if (event.button==2) {app2_elementClickRight(objId);}
	clickCancel();
}

function app2_elementClickLeft(objId) {
	var element=document.getElementById(objId);

	if (element.dataset.controltyp==255) {
		//Gruppe
		//-------------------
		if (event.shiftKey) {	//alle Elemente dieser Gruppe selektieren/deselektieren
			app2_elementSelect(objId,2);
			if (app2_elementIsSelected(objId)) {
				app2_groupSelect(objId,1);
			} else {
				app2_groupSelect(objId,0);
			}
		} else {	
			app2_elementDragStart(objId);
		}

	} else {
		//Element
		//-------------------
		if (event.shiftKey) {
			app2_elementSelect(objId,2);
		} else {
			app2_elementDragStart(objId);
		}

	}

}

function app2_elementClickRight(objId) {
	var element=document.getElementById(objId);

	apps_contextMenu=new class_contextMenu(app2_winId);
	if (element.dataset.controltyp==255) {
		//Gruppe
		apps_contextMenu.addItem("Gruppe bearbeiten","controlClickLeft('"+objId+"');");
		apps_contextMenu.addItem("Gruppe auflösen","app2_uncreateGroup('"+objId+"');");
		apps_contextMenu.addItem("Gruppe duplizieren","app2_elementGroupDuplicate('"+objId+"');");
		apps_contextMenu.addItem("Gruppe löschen","jsConfirm('Soll diese Gruppe (einschließlich Inhalt) wirklich gelöscht werden?','app2_elementGroupDelete($"+objId+"$);','','Löschen');");
	} else {
		//Element
		apps_contextMenu.addItem("Visuelement bearbeiten","controlClickLeft('"+objId+"');");
		apps_contextMenu.addItem("Visuelement duplizieren","app2_elementDuplicate('"+objId+"');");
		apps_contextMenu.addItem("Visuelement löschen","jsConfirm('Soll dieses Visuelement wirklich gelöscht werden?','app2_elementDelete($"+objId+"$);','','Löschen');");
	}
	app2_contextMenuSelection();
	apps_contextMenu.show();
}

function app2_contextMenuSelection() {
	var selectedCount=document.getElementById(app2_winId+"-page").querySelectorAll("[data-elementselected='1']").length;
	var bufferCount=document.getElementById(app2_winId+"-global").dataset.copybuffer.split(";").length-1;
	apps_contextMenu.addHr();
	apps_contextMenu.addItem("Alles auswählen","app2_elementSelectAll();");
	if (selectedCount>0) {
		apps_contextMenu.addItem("Auswahl umkehren","app2_elementSelectInvert();");
		apps_contextMenu.addItem("Auswahl aufheben","app2_elementSelectNone();");
		apps_contextMenu.addHr();
		apps_contextMenu.addText("Ausgewählte Visuelemente ("+selectedCount+"):");
		apps_contextMenu.addItem("&rarr; Merken","app2_elementSelectedToBuffer();");
		apps_contextMenu.addItem("&rarr; Gruppieren","app2_createGroup();");

		apps_contextMenu.addItem("&rarr; Schützen","app2_elementLayerSelected(1);",1);
		apps_contextMenu.addVr();
		apps_contextMenu.addItem("Freigeben","app2_elementLayerSelected(0);",1);

		apps_contextMenu.addItem("&rarr; Duplizieren","app2_elementDuplicateSelected();");

		apps_contextMenu.addItem("&rarr; Z-Index +","app2_elementZindexSelected(1);",1);
		apps_contextMenu.addVr();
		apps_contextMenu.addItem("Z-Index -","app2_elementZindexSelected(-1);",1);

		apps_contextMenu.addItem("&rarr; Löschen","jsConfirm('Sollen wirklich alle markierten Visuelemente ("+selectedCount+") gelöscht werden?','app2_elementDeleteSelected();','','Löschen');");
	}
	if (bufferCount>0) {
		apps_contextMenu.addHr();
		apps_contextMenu.addText("Gemerkte Visuelemente ("+bufferCount+"):");
		apps_contextMenu.addItem("&rarr; Auf diese Seite duplizieren","app2_elementPasteBuffered(0);");
		apps_contextMenu.addItem("&rarr; Auf diese Seite verschieben","app2_elementPasteBuffered(1);");
	}
}


//---------------------------------------------
//globale GUI-Events
//---------------------------------------------
function app2_itemPageClick() {
	if (!event) var event=window.event;
	if (event.button==0) {
		if (event.shiftKey) {
			app2_elementSelectRect(2);
		} else {
			app2_elementSelectNone();
			app2_elementSelectRect(2);
		}
	}
	if (event.button==2) {
		if (app2_pageId>0) {
			var pos=app2_getMousePosition();
			apps_contextMenu=new class_contextMenu(app2_winId);
			apps_contextMenu.addItem("Visuelement hinzufügen","app2_pickElementAtCursor("+pos.x+","+pos.y+");");
			app2_contextMenuSelection();
			apps_contextMenu.show();
		}
	}
}

function app2_pickElementAtCursor(x,y) {
	var rasterX=1;
	var rasterY=1;
	if (app2_gridXmode) {rasterX=parseInt(document.getElementById(app2_winId+"-rasterx").value);}
	if (app2_gridYmode) {rasterY=parseInt(document.getElementById(app2_winId+"-rastery").value);}
	x=Math.floor(x/rasterX)*rasterX;
	y=Math.floor(y/rasterY)*rasterY;
	app2_newElement(x,y);
}

function app2_itemPageMouseMove() {
	if (!event) var event=window.event;
	if (app2_elementSelectRectData.mode>0) {
		app2_elementSelectRect(1);
	}
}

function app2_itemPageUnclick() {
	if (app2_elementSelectRectData.mode>0) {app2_elementSelectRect(0);}
}

function app2_getMousePosition() {
	if (!event) var event=window.event;
	var objPageContainer=document.getElementById(app2_winId+"-pagecontainer");
	var objPage=document.getElementById(app2_winId+"-page");
	var x=(event.pageX*(1/app2_zoomFaktor))-parseInt(objPageContainer.offsetLeft)*(1/app2_zoomFaktor)+parseInt(objPage.scrollLeft);
	var y=(event.pageY*(1/app2_zoomFaktor))-parseInt(objPageContainer.offsetTop)*(1/app2_zoomFaktor)+parseInt(objPage.scrollTop);
	x-=5;	//5px abziehen für "visupage"-Border
	y-=5;	//5px abziehen für "visupage"-Border
	return {x,y};
}


/*
============================================================================
Callbacks
============================================================================
*/

function app2_pickPage_callback(senderId) {
	var newPageId=document.getElementById(senderId).dataset.value;
	if (newPageId>0) {
		app2_refreshAll(newPageId);
		//Zum Ursprung scrollen
		document.getElementById(app2_winId+"-page").scrollTop=0;
		document.getElementById(app2_winId+"-page").scrollLeft=0;
	}
}

function app2_editElement_callback(senderId) {
	if (document.getElementById(senderId).dataset.value>0) {
		app2_refreshAll(0);
	}
}


/*
============================================================================
Funktionalität
============================================================================
*/

function app2_newDiv(objParent,id) {
	//erzeugt neues Div und hängt es an parent
	//objParent: Parent-Objekt
	//id: gewünschte ID des neuen DIVs
	var div=document.createElement('div');
	objParent.appendChild(div);
	div.style.position="absolute";
	div.id=id;
	return div;
}

function app2_elementSelectRect(mode) {
	var visuPage=document.getElementById(app2_winId+"-visupage");
	if (visuPage) {
		var pos=app2_getMousePosition();
		if (mode==2) {
			app2_elementSelectRectData.mode=2;
			app2_elementSelectRectData.x1=pos.x;
			app2_elementSelectRectData.y1=pos.y;
			app2_elementSelectRectData.x2=pos.x;
			app2_elementSelectRectData.y2=pos.y;
			render();
		} else if (mode==1) {
			app2_elementSelectRectData.mode=1;
			app2_elementSelectRectData.x2=pos.x;
			app2_elementSelectRectData.y2=pos.y;
			getHits();
			render();			
		} else if (mode==0) {
			app2_elementSelectRectData.mode=0;
			getHits();
			render();
		}
	}

	function getHits() {
		var visuPage=document.getElementById(app2_winId+"-visupage");
		if (visuPage) {
			//absolute Position von visupage ermitteln
			var tmp=visuPage.getBoundingClientRect();
			var offsetX=tmp.left*(1/app2_zoomFaktor)+5;	//5 Pixel für Border
			var offsetY=tmp.top*(1/app2_zoomFaktor)+5;	//5 Pixel für Border
	
			//Box
			var x1 = Math.min(app2_elementSelectRectData.x1,app2_elementSelectRectData.x2);
			var x2 = Math.max(app2_elementSelectRectData.x1,app2_elementSelectRectData.x2);
			var y1 = Math.min(app2_elementSelectRectData.y1,app2_elementSelectRectData.y2);
			var y2 = Math.max(app2_elementSelectRectData.y1,app2_elementSelectRectData.y2);
	
			var elements=visuPage.querySelectorAll("[data-elementselected='0']");
			for (var t=0; t<elements.length; t++) {
				var tmp=elements[t].getBoundingClientRect();
				var x=(tmp.left*(1/app2_zoomFaktor))-offsetX;
				var y=(tmp.top*(1/app2_zoomFaktor))-offsetY;
				var w=tmp.width*(1/app2_zoomFaktor);
				var h=tmp.height*(1/app2_zoomFaktor);
	
				if ((x>=x1 && x<=x2 && y>=y1 && y<=y2) && ((x+w)>=x1 && (x+w)<=x2 && (y+h)>=y1 && (y+h)<=y2)) {
					if (app2_elementSelectRectData.mode==0) {
						app2_elementSelect(elements[t].id,1); 
					} else if (elements[t].dataset.layer==0 || elements[t].dataset.layer==app2_layerMode) {
						elements[t].style.boxShadow=app2_el_shadowSelected;
						document.getElementById(elements[t].id+"-list").style.background=apps_colorSelected;
					}
				} else {
					elements[t].style.boxShadow=elements[t].dataset.shadowbuffer;
					document.getElementById(elements[t].id+"-list").style.background="";
				}
	
			}
		}
	}

	function render() {
		var visuPage=document.getElementById(app2_winId+"-visupage");
		if (visuPage) {
			if (app2_elementSelectRectData.mode>0) {
				var tmp=document.getElementById(app2_winId+"-selectrect");
				if (!tmp) {
					tmp=app2_newDiv(visuPage,app2_winId+"-selectrect");
					tmp.className="app2_elementSelectRect";
				}		
				var x1 = Math.min(app2_elementSelectRectData.x1,app2_elementSelectRectData.x2);
				var x2 = Math.max(app2_elementSelectRectData.x1,app2_elementSelectRectData.x2);
				var y1 = Math.min(app2_elementSelectRectData.y1,app2_elementSelectRectData.y2);
				var y2 = Math.max(app2_elementSelectRectData.y1,app2_elementSelectRectData.y2);
				tmp.style.left=x1+"px";
				tmp.style.top=y1+"px";
				tmp.style.width=(x2-x1)+"px";
				tmp.style.height=(y2-y1)+"px";
			} else {
				var tmp=document.getElementById(app2_winId+"-selectrect");
				if (tmp) {tmp.parentNode.removeChild(tmp);}
			}
		}
	}
}

function app2_elementSelect(objId,mode) {
	var element=document.getElementById(objId);
	if (element) {
		if (element.dataset.layer==0 || element.dataset.layer==app2_layerMode) {
			if (mode==2) {
				if (element.dataset.elementselected==0) {mode=1;} else {mode=0;}
			}
			if (mode==1) {
				element.style.boxShadow=app2_el_shadowSelected;
				document.getElementById(objId+"-list").style.background=apps_colorSelected;
				element.dataset.elementselected=1;
			} else if (mode==0) {
				element.style.boxShadow=element.dataset.shadowbuffer;
				document.getElementById(objId+"-list").style.background="";
				element.dataset.elementselected=0;
			}
		}
	}
}

function app2_elementSelectAll() {
	var page=document.getElementById(app2_winId+"-visupage");
	if (page) {
		var element=page.querySelectorAll("[data-elementselected='0']");
		for (var t=0; t<element.length; t++) {
			app2_elementSelect(element[t].id,1);
		}
	}
}

function app2_elementSelectInvert() {
	var page=document.getElementById(app2_winId+"-visupage");
	if (page) {
		var element=page.querySelectorAll("[data-elementselected]");
		for (var t=0; t<element.length; t++) {
			if (element[t].dataset.elementselected==1) {
				app2_elementSelect(element[t].id,0);
			} else {
				app2_elementSelect(element[t].id,1);
			}
		}
	}
}

function app2_elementSelectNone() {
	var page=document.getElementById(app2_winId+"-visupage");
	if (page) {
		var element=page.querySelectorAll("[data-elementselected='1']");
		for (var t=0; t<element.length; t++) {
			app2_elementSelect(element[t].id,0);
		}
	}
}

function app2_elementIsSelected(objId) {
	var element=document.getElementById(objId);
	if (element) {
		if (element.dataset.elementselected==1) {return true;}
		if (element.dataset.elementselected==0) {return false;}
	}
}

function app2_elementCountSelected() {
	var element=document.getElementById(app2_winId+"-visupage").querySelectorAll("[data-elementselected='1']");
	return element.length;
}

function app2_groupSelect(objId,mode) {
	var element=document.getElementById(objId);
	if (element) {
		if (element.dataset.controltyp=="255") {
			var elements=document.getElementById(app2_winId+"-visupage").querySelectorAll("[data-groupid='"+element.dataset.elementid+"']");
			for (var t=0; t<elements.length; t++) {
				app2_elementSelect(elements[t].id,mode);
			}
		}
	}
}


function app2_elementAddBorder(obj,enabled) {
	if (enabled) {
		if (obj.dataset.layer==0 || obj.dataset.layer==app2_layerMode) {var color=app2_elementOutlineColor;} else {var color=app2_elementOutlineColorDisabled;}
	} else {
		var color=app2_elementOutlineColorDisabled;
	}
	if (!app2_previewMode) {
		obj.style.outline=color+" solid 1px";
		obj.style.outlineOffset="-1px";
	}
}

function app2_groupAddBorder(obj,enabled) {
	if (enabled) {
		if (obj.dataset.layer==0 || obj.dataset.layer==app2_layerMode) {var color=app2_elementOutlineColor;} else {var color=app2_elementOutlineColorDisabled;}
	} else {
		var color=app2_elementOutlineColorDisabled;
	}
	if (!app2_previewMode) {
		obj.style.outline=color+" dashed 1px";
		obj.style.outlineOffset="-1px";
	}
}

function app2_groupBoundingBoxes() {
	//Boundingboxen aller Gruppen ermitteln und anzeigen

	var xpos,ypos,xsize,ysize;

	//absolute Position von visupage ermitteln (als Offset für getBoundingClientRect() notwendig)
	var tmp=document.getElementById(app2_winId+"-visupage").getBoundingClientRect();
	var offsetX=tmp.left*(1/app2_zoomFaktor)+5;	//5 Pixel für Border
	var offsetY=tmp.top*(1/app2_zoomFaktor)+5;	//5 Pixel für Border

	//Gruppen suchen
	var group=document.getElementById(app2_winId+"-visupage").querySelectorAll("[data-controltyp='255']");
	for (var t=0; t<group.length; t++) {

		var x1=null;
		var y1=null;
		var x2=0;
		var y2=0;
		
		//Elemente der Gruppe suchen
		var item=document.getElementById(app2_winId+"-visupage").querySelectorAll("[data-groupid='"+group[t].dataset.elementid+"']");
		for (var tt=0; tt<item.length; tt++) {
			
			//min/max... und speichern! (dataset) => Dragging...
			xpos=parseInt(item[tt].style.left.replace("px",""));
			ypos=parseInt(item[tt].style.top.replace("px",""));
			xsize=parseInt(item[tt].offsetWidth);
			ysize=parseInt(item[tt].offsetHeight);

			tmp=item[tt].getBoundingClientRect();
			xpos=(tmp.left*(1/app2_zoomFaktor))-offsetX;
			ypos=(tmp.top*(1/app2_zoomFaktor))-offsetY;
			xsize=tmp.width*(1/app2_zoomFaktor);
			ysize=tmp.height*(1/app2_zoomFaktor);
			
			if (xpos<x1 || x1==null) {x1=xpos;}
			if (ypos<y1 || y1==null) {y1=ypos;}
			if ((xpos+xsize)>x2) {x2=xpos+xsize;}
			if ((ypos+ysize)>y2) {y2=ypos+ysize;}
			
		}

		if (item.length==0) {	//leere Gruppe
			group[t].style.left="0px";
			group[t].style.top="0px";
			group[t].style.width="0px";
			group[t].style.height="0px";
			group[t].style.display="none";			
		} else {
			group[t].style.left=Math.round(x1)+"px";
			group[t].style.top=Math.round(y1)+"px";
			group[t].style.width=Math.round(x2-x1)+"px";
			group[t].style.height=Math.round(y2-y1)+"px";
			group[t].style.display="block";			
		}
		
	}
}

function app2_selectionBoundingBox() {
	//Boundingbox der ausgewählten Elemente berechnen

	var xpos,ypos,xsize,ysize;

	var x1=null;
	var y1=null;
	var x2=0;
	var y2=0;

	//absolute Position von visupage ermitteln (als Offset für getBoundingClientRect() notwendig)
	var tmp=document.getElementById(app2_winId+"-visupage").getBoundingClientRect();
	var offsetX=tmp.left*(1/app2_zoomFaktor)+5;	//5 Pixel für Border
	var offsetY=tmp.top*(1/app2_zoomFaktor)+5;	//5 Pixel für Border

	var elements=document.getElementById(app2_winId+"-visupage").querySelectorAll("[data-elementselected='1']");
	for (var t=0; t<elements.length; t++) {
		if (elements[t].dataset.controltyp!=255) { //keine Gruppenelemente (Gruppe könnte leer sein - dann stimmt die Position ja nicht)
			//min/max... und speichern! (dataset) => Dragging...
			xpos=parseInt(elements[t].style.left.replace("px",""));
			ypos=parseInt(elements[t].style.top.replace("px",""));
			xsize=parseInt(elements[t].offsetWidth);
			ysize=parseInt(elements[t].offsetHeight);

			tmp=elements[t].getBoundingClientRect();
			xpos=(tmp.left*(1/app2_zoomFaktor))-offsetX;
			ypos=(tmp.top*(1/app2_zoomFaktor))-offsetY;
			xsize=tmp.width*(1/app2_zoomFaktor);
			ysize=tmp.height*(1/app2_zoomFaktor);

			if (xpos<x1 || x1==null) {x1=xpos;}
			if (ypos<y1 || y1==null) {y1=ypos;}
			if ((xpos+xsize)>x2) {x2=xpos+xsize;}
			if ((ypos+ysize)>y2) {y2=ypos+ysize;}
		}
	}

	return {
		x:Math.round(x1),
		y:Math.round(y1),
		width:Math.round(x2-x1),
		height:Math.round(y2-y1)
	};
		
}

//---------------------------------------------
//diverse Funktionen
//---------------------------------------------

function app2_elementDuplicate(objId) {
	var obj=document.getElementById(objId);
	if (obj) {
		ajax("pasteElements",2,app2_winId,app2_pageId,obj.dataset.elementid);
	}
}
function app2_elementDelete(objId) {
	var obj=document.getElementById(objId);
	if (obj) {
		ajax("deleteElements",2,app2_winId,app2_pageId,obj.dataset.elementid);
	}
}
function app2_elementGroupDuplicate(objId) {
	var obj=document.getElementById(objId);
	if (obj) {
		data=obj.dataset.elementid+";";
		var element=document.getElementById(app2_winId+"-visupage").querySelectorAll("[data-groupid='"+obj.dataset.elementid+"']");
		for (var t=0; t<element.length; t++) {
			if (element[t].dataset.layer==0 || element[t].dataset.layer==app2_layerMode) {
				data+=element[t].dataset.elementid+";";
			}
		}
		if (data!="") {ajax("pasteElements",2,app2_winId,app2_pageId,data);}
	}
}
function app2_elementGroupDelete(objId) {
	var obj=document.getElementById(objId);
	if (obj) {
		data=obj.dataset.elementid+";";
		var element=document.getElementById(app2_winId+"-visupage").querySelectorAll("[data-groupid='"+obj.dataset.elementid+"']");
		for (var t=0; t<element.length; t++) {
			if (element[t].dataset.layer==0 || element[t].dataset.layer==app2_layerMode) {
				data+=element[t].dataset.elementid+";";
			}
		}
		if (data!="") {ajax("deleteElements",2,app2_winId,app2_pageId,data);}
	}
}

function app2_createGroup() {
	var data="";
	var element=document.getElementById(app2_winId+"-visupage").querySelectorAll("[data-elementselected='1']");
	if (element.length>0) {	//mindestens 1 Elemente selektiert?
		for (var t=0; t<element.length; t++) {data+=element[t].dataset.elementid+";";}
		if (data!="") {
			ajax("createGroup",2,app2_winId,app2_pageId,data);
		}
	}
}

function app2_uncreateGroup(objId) {
	var element=document.getElementById(objId);
	if (element.dataset.controltyp==255) {
		ajax("uncreateGroup",2,app2_winId,app2_pageId,element.dataset.elementid);
	}
}

function app2_groupCollapse(objId,mode) {
	var element=document.getElementById(objId);
	if (element) {
		if (mode==2) {
			if (element.dataset.groupexpanded==0) {mode=1;} else {mode=0;}
		}
		if (mode==1) {
			element.style.display="block";
			document.getElementById(objId+"-img").src="../shared/img/folder1b.png";
			element.dataset.groupexpanded=1;
		} else if (mode==0) {
			element.style.display="none";
			document.getElementById(objId+"-img").src="../shared/img/folder1.png";
			element.dataset.groupexpanded=0;
		}
	}
}

function app2_moveElementsByKeyboard() {
	if (app2_pageId>0) {
		app2_elementsKeyboardMoveWinId=openWindow(1013,AJAX_SEPARATOR1+"1013"+AJAX_SEPARATOR1+"-1"+AJAX_SEPARATOR1+AJAX_SEPARATOR1+app2_pageId);
		app2_elementDragStart(null,true);
	}
}

function app2_moveElementsByKeyboardKeyEvent() {
	if (!event) var event=window.event;
	if (event.keyCode==37) {app2_elementDragMove(null,true,-1,0);} 	//Cursor links
	if (event.keyCode==38) {app2_elementDragMove(null,true,0,-1);} 	//Cursor hoch
	if (event.keyCode==39) {app2_elementDragMove(null,true,1,0);} 	//Cursor rechts
	if (event.keyCode==40) {app2_elementDragMove(null,true,0,1);} 	//Cursor runter
	if (event.keyCode==27) {app2_moveElementsByKeyboardCancel(app2_elementsKeyboardMoveWinId);}	//ESC
	if (event.keyCode==13) {app2_moveElementsByKeyboardSave(app2_elementsKeyboardMoveWinId);} 	//Enter
	event.cancelBubble=true;
	if (event.stopPropagation) {event.stopPropagation();}
	event.preventDefault();
	return false;
}

function app2_moveElementsByKeyboardCancel(winId) {
	app2_elementDragEnd(null,true,false);
	window.removeEventListener("keydown",app2_moveElementsByKeyboardKeyEvent,false);
	closeWindow(winId);
	app2_refreshAll(0);
}

function app2_moveElementsByKeyboardSave(winId) {
	app2_elementDragEnd(null,true,true);
	window.removeEventListener("keydown",app2_moveElementsByKeyboardKeyEvent,false);
	closeWindow(winId);
}

function app2_newElement(x,y) {
	if (app2_pageId>0) {
		openWindow(1002,AJAX_SEPARATOR1+"1002"+AJAX_SEPARATOR1+"-1"+AJAX_SEPARATOR1+AJAX_SEPARATOR1+app2_pageId+AJAX_SEPARATOR1+x+AJAX_SEPARATOR1+y);
	}
}

function app2_setLayerMode() {
	//(aktuell nur toggeln zw. 0 und 1)
	if (app2_layerMode==0) {
		app2_layerMode=1;
		document.getElementById(app2_winId+"-layer").style.background="#80e000";
		document.getElementById(app2_winId+"-layer").innerHTML="<img src='../shared/img/lock1b.png' width='16' height='16' valign='middle' style='margin:0; padding-left:2px;' draggable='false'>";
	} else {
		app2_layerMode=0;
		document.getElementById(app2_winId+"-layer").style.background="";
		document.getElementById(app2_winId+"-layer").innerHTML="<img src='../shared/img/lock1.png' width='16' height='16' valign='middle' style='margin:0; padding-left:2px;' draggable='false'>";
	}
	app2_refreshAll(0);
}

function app2_previewModeToggle() {
	if (app2_previewMode) {
		app2_previewMode=false;
		document.getElementById(app2_winId+"-preview").style.background="";
		document.getElementById(app2_winId+"-previewValue").style.display="none";
	} else {
		app2_previewMode=true;
		document.getElementById(app2_winId+"-preview").style.background="#80e000";
		document.getElementById(app2_winId+"-previewValue").style.display="inline-block";
	}
	app2_refreshAll(0);
}

function app2_previewChangeValue(obj) {
	app2_previewValue=stringCleanup(obj.value);
	if (app2_previewMode) {
		app2_refreshAll(0);
	}
}

function app2_previewGetCSS(n) {
	if (app2_previewMode) {
		return pS(n,app2_previewValue);
	} else {
		return n;
	}
}

function app2_gridXtoggle() {
	if (app2_gridXmode) {
		app2_gridXmode=false;
		document.getElementById(app2_winId+"-rasterx").style.color="#a0a0a0";
	} else {
		app2_gridXmode=true;
		document.getElementById(app2_winId+"-rasterx").style.color="#000000";
	}
}

function app2_gridYtoggle() {
	if (app2_gridYmode) {
		app2_gridYmode=false;
		document.getElementById(app2_winId+"-rastery").style.color="#a0a0a0";
	} else {
		app2_gridYmode=true;
		document.getElementById(app2_winId+"-rastery").style.color="#000000";
	}
}


//---------------------------------------------
//Selektion und Copy/Paste
//---------------------------------------------

function app2_elementLayerSelected(mode) {
	//markierte Objekte: layer zuweisen
	var data="";
	var page=document.getElementById(app2_winId+"-visupage");
	if (page) {
		var element=page.querySelectorAll("[data-elementselected='1']");
		for (var t=0; t<element.length; t++) {data+=element[t].dataset.elementid+";";}
		if (data!="") {
			ajax("layerElements",2,app2_winId,app2_pageId,data+AJAX_SEPARATOR1+mode);
		}
	}
}

function app2_elementDuplicateSelected() {
	//markierte Elemente auf aktueller Seite duplizieren
	var data="";
	var element=document.getElementById(app2_winId+"-visupage").querySelectorAll("[data-elementselected='1']");
	for (var t=0; t<element.length; t++) {data+=element[t].dataset.elementid+";";}
	if (data!="") {ajax("pasteElements",2,app2_winId,app2_pageId,data);}
}

function app2_elementZindexSelected(mode) {
	//erhöht/erniedrigt den Zindex aller markierten Objekte
	var data="";
	var page=document.getElementById(app2_winId+"-visupage");
	if (page) {
		var element=page.querySelectorAll("[data-elementselected='1']");
		for (var t=0; t<element.length; t++) {data+=element[t].dataset.elementid+";";}
		if (data!="") {
			if (mode==1) {ajax("incElementsZindex",2,app2_winId,app2_pageId,data);}
			if (mode==-1) {ajax("decElementsZindex",2,app2_winId,app2_pageId,data);}
		}
	}
}

function app2_elementDeleteSelected() {
	//markierte Elemente löschen
	var data="";
	var element=document.getElementById(app2_winId+"-visupage").querySelectorAll("[data-elementselected='1']");
	for (var t=0; t<element.length; t++) {data+=element[t].dataset.elementid+";";}
	if (data!="") {
		ajax("deleteElements",2,app2_winId,app2_pageId,data);
	}
}

function app2_elementSelectedToBuffer() {
	//markierte Elemente merken
	var data="";
	var element=document.getElementById(app2_winId+"-visupage").querySelectorAll("[data-elementselected='1']");
	for (var t=0; t<element.length; t++) {data+=element[t].dataset.elementid+";";}
	document.getElementById(app2_winId+"-global").dataset.copybuffer=data;
}

function app2_elementPasteBuffered(mode) {
	//gemerkte Elemente auf aktuelle Seite einfügen
	var data=document.getElementById(app2_winId+"-global").dataset.copybuffer;
	if (data!="") {
		if (mode==1) {
			ajax("moveElements",2,app2_winId,app2_pageId,data);
		} else {
			ajax("pasteElements",2,app2_winId,app2_pageId,data);
		}
	}
	document.getElementById(app2_winId+"-global").dataset.copybuffer="";
}


/*
============================================================================
Visuelement erstellen
============================================================================
*/

function class_app2_VE() {

	var that=this;
	this.element;
	this.item;

	var visupage=document.getElementById(app2_winId+"-visupage");
	var menuItem=null;
	var menuItemCaption="";

	this.addGroup=function() {
		if (that.element.pageid==app2_pageId) {	//Element der aktuellen Seite
			//Visuelement (Gruppenelement)
			that.item=app2_newDiv(visupage,app2_winId+"-element-"+that.element.id);
			that.item.dataset.elementid=that.element.id;
			that.item.dataset.controltyp=that.element.controltyp;
			that.item.dataset.groupid="0";
			that.item.dataset.shadowbuffer="none";
			that.item.dataset.elementselected=0;		
			that.item.dataset.layer=that.element.layer;
			if (that.element.layer==0 || that.element.layer==app2_layerMode) {
				that.item.className="app2_element";
				that.item.setAttribute("onMouseDown","app2_elementClick('"+that.item.id+"');");
				//Control 1002
				that.item.dataset.type=1002;
				that.item.dataset.value=that.element.id;
				that.item.dataset.callback="app2_editElement_callback('"+that.item.id+"');";
			} else {
				that.item.style.cursor="not-allowed";
			}	
			app2_groupAddBorder(that.item,true);
	
			//Menuelement (Gruppenelement)
			menuItem=createNewDiv(app2_winId+"-menu",app2_winId+"-element-"+that.element.id+"-list");
			menuItem.className="controlListItem";
			menuItem.style.display="block";
			menuItem.style.color="#0000ff";
			menuItem.innerHTML="<img id='"+app2_winId+"-element-"+that.element.id+"-group-img' src='../shared/img/folder1.png' width='12' height='12' onClick='app2_groupCollapse(\""+app2_winId+"-element-"+that.element.id+"-group\",2);' valign='middle' style='margin:0; padding-left:2px;' draggable='false'> <b>"+that.element.name+"</b> <span class='id' style='background:#8080ff;'>"+that.element.id+"</span>";		
			if (that.element.layer==0 || that.element.layer==app2_layerMode) {
				menuItem.setAttribute("onMouseDown","app2_elementClick('"+app2_winId+"-element-"+that.element.id+"');");
			} else {
				menuItem.style.opacity="0.5";
				menuItem.style.cursor="not-allowed";
			}
		
			//Menuelement (Container für Elemente dieser Gruppe)
			var groupItem=createNewDiv(app2_winId+"-menu",app2_winId+"-element-"+that.element.id+"-group");
			groupItem.style.display="none";
			groupItem.dataset.groupexpanded=0;
		}
	}
	
	this.addElement=function() {
		if (that.element.pageid==app2_pageId) {	//Element der aktuellen Seite
			//Visuelement
			that.item=app2_newDiv(visupage,app2_winId+"-element-"+that.element.id);
			that.item.dataset.elementid=that.element.id;
			that.item.dataset.controltyp=that.element.controltyp;
			that.item.dataset.groupid=that.element.groupid;
			that.item.dataset.shadowbuffer="none";
			that.item.dataset.elementselected=0;
			that.item.dataset.layer=that.element.layer;
			if (that.element.layer==0 || that.element.layer==app2_layerMode) {
				that.item.setAttribute("onMouseDown","app2_elementClick('"+that.item.id+"');");
				that.item.className="app2_element";
				//Control 1002
				that.item.dataset.type=1002;
				that.item.dataset.value=that.element.id;
				that.item.dataset.callback="app2_editElement_callback('"+that.item.id+"');";
			} else {
				that.item.setAttribute("onMouseDown","clickCancel();");	//verhindert die Anzeige des allgemeinen Kontextmenüs beim Klick auf das Visuelement
				that.item.style.cursor="not-allowed";
			}
			app2_elementAddBorder(that.item,true);

			//Menuelement
			if (that.element.groupid>0) {
				//Element einer Gruppe -> in den Gruppencontainer einhängen
				menuItem=createNewDiv(app2_winId+"-element-"+that.element.groupid+"-group",app2_winId+"-element-"+that.element.id+"-list");
				menuItem.className="controlListItem";
				menuItem.style.display="block";
				menuItem.style.paddingLeft="17px";
				if (that.element.layer==0 || that.element.layer==app2_layerMode) {
					menuItem.setAttribute("onMouseDown","app2_elementClick('"+that.item.id+"');");
				} else {
					menuItem.style.opacity="0.5";
					menuItem.style.cursor="not-allowed";
				}
			} else {
				//Element ohne Gruppe
				menuItem=createNewDiv(app2_winId+"-menu",app2_winId+"-element-"+that.element.id+"-list");
				menuItem.className="controlListItem";
				menuItem.style.display="block";
				if (that.element.layer==0 || that.element.layer==app2_layerMode) {
					menuItem.setAttribute("onMouseDown","app2_elementClick('"+that.item.id+"');");
				} else {
					menuItem.style.opacity="0.5";
					menuItem.style.cursor="not-allowed";
				}
			}
		} else {	//Inkludieres Element
			//Visuelement
			that.item=app2_newDiv(visupage,app2_winId+"-element-"+that.element.id);
			that.item.style.cursor="not-allowed";
			app2_elementAddBorder(that.item,false);
		}		
	}
		
	this.createElement=function(property) {
		if (app2_previewMode) {
			var caption=pS(that.element.text,app2_previewValue);
		} else {
			var caption=that.element.text;
		}

		var menuCaption="<b>"+that.element.name+"</b> <span class='id'>"+that.element.id+"</span>";
		
		if (that.element.controltyp==1) {
			control1_preview(that.item,caption);
			setMenuItemCaption(menuCaption+" "+that.element.menutext);
			
		} else if (that.element.controltyp==10) {
			control10_preview(that.item,caption);
			setMenuItemCaption(menuCaption+" <span class='app2_pseudoElement'>TASTATUREINGABE</span> "+that.element.menutext);
			
		} else if (that.element.controltyp==11) {
			control11_preview(that.item,caption);
			setMenuItemCaption(menuCaption+" <span class='app2_pseudoElement'>DREHREGLER</span> "+that.element.menutext);
			
		} else if (that.element.controltyp==12) {
			control12_preview(that.item,caption);
			setMenuItemCaption(menuCaption+" <span class='app2_pseudoElement'>DIMMER/RGB/HSV</span> "+that.element.menutext);
			
		} else if (that.element.controltyp==13) {
			control13_preview(that.item,caption);
			setMenuItemCaption(menuCaption+" <span class='app2_pseudoElement'>SCHIEBEREGLER</span> "+that.element.menutext);
			
		} else if (that.element.controltyp==14) {
			control14_preview(that.item,property.p0);
			setMenuItemCaption(menuCaption+" <span class='app2_pseudoElement'>LISTE/TABELLE</span>");
			
		} else if (that.element.controltyp==15) {
			control15_preview(that.item,caption);
			setMenuItemCaption(menuCaption+" <span class='app2_pseudoElement'>FARBAUSWAHL</span> "+that.element.menutext);
			
		} else if (that.element.controltyp==16) {
			control16_preview(that.item,property.p0,caption);
			setMenuItemCaption(menuCaption+" <span class='app2_pseudoElement'>NOTIZEN/SKIZZE</span> "+that.element.menutext);
			
		} else if (that.element.controltyp==17) {
			control17_preview(that.item,caption);
			setMenuItemCaption(menuCaption+" <span class='app2_pseudoElement'>ANALOGUHR</span> "+that.element.menutext);
			
		} else if (that.element.controltyp==20) {
			control20_preview(that.item,property.p0,property.p1,property.p2);
			setMenuItemCaption(menuCaption+" <span class='app2_pseudoElement'>KAMERABILD/-ARCHIV</span> "+property.p3);
			
		} else if (that.element.controltyp==21) {
			control21_preview(that.item,property.p0);
			setMenuItemCaption(menuCaption+" <span class='app2_pseudoElement'>DIAGRAMM</span> "+property.p1);
			
		} else if (that.element.controltyp==22) {
			control22_preview(that.item,property.p0,property.p1);
			setMenuItemCaption(menuCaption+" <span class='app2_pseudoElement'>ZEITSCHALTUHR</span> "+property.p2);
			
		} else if (that.element.controltyp==23) {
			control23_preview(that.item,property.p0,property.p1);
			setMenuItemCaption(menuCaption+" <span class='app2_pseudoElement'>ANWESENHEITSSIMULATION</span> "+property.p2);
			
		} else if (that.element.controltyp==24) {
			control24_preview(that.item);
			setMenuItemCaption(menuCaption+" <span class='app2_pseudoElement'>CODESCHLOSS</span>");
			
		} else if (that.element.controltyp==25) {
			control25_preview(that.item,property.p0,property.p1);
			setMenuItemCaption(menuCaption+" <span class='app2_pseudoElement'>MELDUNGSARCHIV</span> "+property.p2);

		} else if (that.element.controltyp==26) {
			control26_preview(that.item,property.p0,property.p1);
			setMenuItemCaption(menuCaption+" <span class='app2_pseudoElement'>DATENARCHIV</span> "+property.p2);

		} else if (that.element.controltyp==27) {
			control27_preview(that.item,caption);
			setMenuItemCaption(menuCaption+" <span class='app2_pseudoElement'>RUNDINSTRUMENT</span> "+that.element.menutext);

		} else if (that.element.controltyp==28) {
			control28_preview(that.item,property.p0,property.p1);
			setMenuItemCaption(menuCaption+" <span class='app2_pseudoElement'>ANRUFARCHIV</span> "+property.p2);

		} else if (that.element.controltyp==29) {
			control29_preview(that.item);
			setMenuItemCaption(menuCaption+" <span class='app2_pseudoElement'>BILD-URL/WEBSEITE</span>");
			
		} else if (that.element.controltyp==30) {
			control30_preview(that.item);
			setMenuItemCaption(menuCaption+" <span class='app2_pseudoElement'>TON-URL</span>");
			
		} else if (that.element.controltyp==31) {
			control31_preview(that.item);
			setMenuItemCaption(menuCaption+" <span class='app2_pseudoElement'>SPRACHAUSGABE</span>");

		} else if (that.element.controltyp==32) {
			control32_preview(that.item,property.p0,property.p1);
			setMenuItemCaption(menuCaption+" <span class='app2_pseudoElement'>TERMINSCHALTUHR</span> "+property.p2);

		} else if (that.element.controltyp==33) {
			control33_preview(that.item,caption);
			setMenuItemCaption(menuCaption+" <span class='app2_pseudoElement'>TOUCHPAD</span> "+that.element.menutext);
		}
		that.item.dataset.shadowbuffer=that.item.style.boxShadow;
	}
	
	this.appendToMenuItemCaption=function(n) {
		menuItemCaption+=n;
	}
	
	function setMenuItemCaption(n) {
		if (menuItem) {menuItem.innerHTML=n+menuItemCaption;}
	}

}


